//use Memento's LIS3DH accelerometer to measure camera rotation
#define ACCEL_I2C_ADDR 0x19
#define ACCEL_REG_WHOAMI 0x0F //read this register to check that accelerometer is there (should return 0x33)
#define ACCEL_REG_CTRL1 0x20
#define ACCEL_REG_CTRL4 0x23
#define ACCEL_REG_OUT_X 0x28 //read X axis acceleration (int16 value)
#define ACCEL_REG_OUT_Y 0x2A //read Y axis acceleration (int16 value)
#define ACCEL_REG_OUT_Z 0x2C //read Z axis acceleration (int16 value)

byte setupAccel()
{
  Wire.begin();
  if(accelReadByte(ACCEL_REG_WHOAMI) != 0x33) { return false; } //read WHOAMI register to check if accel is there - should return 0x33
  accelWriteByte(ACCEL_REG_CTRL1, 0b01110111); //enable all axes, normal mode, 400Hz update rate
  accelWriteByte(ACCEL_REG_CTRL4, 0b10111000); //high res & enable BDU
  return true;
}

//read 1 byte from accelerometer register
uint8_t accelReadByte(uint8_t reg)
{
  //tell accelerometer which register we want to read from
  Wire.beginTransmission(ACCEL_I2C_ADDR);
  Wire.write(reg);
  Wire.endTransmission();
  //request 1 byte from accelerometer
  Wire.requestFrom(ACCEL_I2C_ADDR, 1);
  if(Wire.available()) { return Wire.read(); }
}

//write 1 byte to accelerometer register
void accelWriteByte(uint8_t reg, uint8_t data)
{
  Wire.beginTransmission(ACCEL_I2C_ADDR);
  Wire.write(reg);
  Wire.write(data);
  Wire.endTransmission();
}

//read 2-byte int from two consecutive accelerometer registers
int16_t accelReadInt(uint8_t reg)
{
  return (accelReadByte(reg + 1) << 8) | accelReadByte(reg);
}

//calculate G force value from raw accelerometer readings (1G = 9.8m/s)
//to mitigate noise, readings are low-pass-filtered by taking a weighted average with the previous reading
//axis numbers: 0=x, 1=y, 2=z
#define ACCEL_SMOOTH 0.75 //amount of smoothing (0-1, higher = more smoothing)
float accel_values[3] = {0, 0, 0};
float accelReadG(uint8_t axis)
{
  float value = 48 * (float(accelReadInt(ACCEL_REG_OUT_X + axis * 2)) / 64000);
  accel_values[axis] = (accel_values[axis] * ACCEL_SMOOTH) + (value * (1.0 - ACCEL_SMOOTH));
  return accel_values[axis];
}