byte setupCamera()
{
  //initial camera hardware setup
	Wire.begin();
	camera_config.ledc_channel = LEDC_CHANNEL_0;
	camera_config.ledc_timer = LEDC_TIMER_0;
	camera_config.pin_d0 = Y2_GPIO_NUM;
	camera_config.pin_d1 = Y3_GPIO_NUM;
	camera_config.pin_d2 = Y4_GPIO_NUM;
	camera_config.pin_d3 = Y5_GPIO_NUM;
	camera_config.pin_d4 = Y6_GPIO_NUM;
	camera_config.pin_d5 = Y7_GPIO_NUM;
	camera_config.pin_d6 = Y8_GPIO_NUM;
	camera_config.pin_d7 = Y9_GPIO_NUM;
	camera_config.pin_xclk = XCLK_GPIO_NUM;
	camera_config.pin_pclk = PCLK_GPIO_NUM;
	camera_config.pin_vsync = VSYNC_GPIO_NUM;
	camera_config.pin_href = HREF_GPIO_NUM;
	camera_config.pin_sccb_sda = -1;
	camera_config.pin_sccb_scl = -1;
	camera_config.sccb_i2c_port = 0; // use the 'first' i2c port
	camera_config.pin_pwdn = PWDN_GPIO_NUM;
	camera_config.pin_reset = RESET_GPIO_NUM;
	camera_config.xclk_freq_hz = 20000000;
	camera_config.grab_mode = CAMERA_GRAB_LATEST;//CAMERA_GRAB_WHEN_EMPTY;
	camera_config.fb_location = CAMERA_FB_IN_PSRAM;
	camera_config.fb_count = 2;
	camera_config.pixel_format = PIXFORMAT_RGB565; //565 pixel format for immediately displaying to LCD
  camera_config.frame_size = FRAMESIZE_QVGA; //capture photos at 320x240, the best resolution
  if(esp_camera_init(&camera_config) != ESP_OK) { return false; }
	
  //initial image processing and exposure parameters setup
	camera = esp_camera_sensor_get();
	camera->set_hmirror(camera, 0); //don't mirror image horizontally
  camera->set_vflip(camera, 1); //flip image vertically (due to sensor orientation)
	camera->set_bpc(camera, 1); //enable black pixel correction
	camera->set_wpc(camera, 1); //enable white pixel correction
	camera->set_brightness(camera, 0); //range: -3 to 3
	camera->set_contrast(camera, 0); //range -3 to 3
	camera->set_saturation(camera, 0); //range -4 to 4
	camera->set_sharpness(camera, 1); //range: -3 to 3
	camera->set_denoise(camera, 0); //disable denoising for maximum crust (denoising range is 0-8)
	camera->set_ae_level(camera, 0); //range: -5 to 5 (auto exposure compensation, only in effect when set_exposure_ctrl and/or set_gain_ctrl are set to 1 (auto))
	camera->set_raw_gma(camera, 1); //[???] enable RAW gamma
	camera->set_whitebal(camera, 1); //enable white balance
	camera->set_awb_gain(camera, 1); //allow adjusting white balance mode
	camera->set_wb_mode(camera, 0); //set white balance to auto (0=auto, 1=sunny, 2=cloudy, 3=office, 4=home)
	camera->set_exposure_ctrl(camera, 1); //enable auto exposure time (set manual exposure time with set_aec_value() ; range is 0-984)
	camera->set_gain_ctrl(camera, 1); //enable auto gain (set manual gain with set_agc_gain() ; range is 0-64)
  return true;
}

byte setupAutofocus()
{
  //download autofocus program to camera
  camera->set_reg(camera, 0x3000, 0xff, 0x20);
  for(uint16_t i = 0; i < sizeof(OV5640_AF_PROGRAM); i++)
  {
    camera->set_reg(camera, 0x8000 + i, 0xff, OV5640_AF_PROGRAM[i]);
  }
  camera->set_reg(camera, OV5640_CMD_MAIN, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_ACK, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_PARA0, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_PARA1, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_PARA2, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_PARA3, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_PARA4, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_FW_STATUS, 0xff, 0x7f);
  camera->set_reg(camera, 0x3000, 0xff, 0x00);

  //wait for autofocus program to start
  uint32_t timeout = millis() + AUTOFOCUS_TIMEOUT;
  while(getAutofocusStatus() != FW_STATUS_S_IDLE)
  {
    if(millis() >= timeout) { return false; }
  }
  if(!sendAutofocusCommand(AF_RELEASE_FOCUS)) { return false; }

  //on startup, set focus to infinity
  setUIValue(PARAM_FOCUS, params[PARAM_FOCUS].max_ui);
  return true;
}

//send command to autofocus program
void sendAutofocusCommandNoAck(uint8_t cmd)
{
  camera->set_reg(camera, OV5640_CMD_ACK, 0xff, 0x01);
  camera->set_reg(camera, OV5640_CMD_MAIN, 0xff, cmd);
}

//send command to autofocus program and wait for it to reply with an ack
byte sendAutofocusCommand(uint8_t cmd)
{
  sendAutofocusCommandNoAck(cmd);
  uint32_t timeout = millis() + AUTOFOCUS_TIMEOUT;
  while(camera->get_reg(camera, OV5640_CMD_ACK, 0xff) != 0x00)
  {
    if(millis() >= timeout) { return false; }
  }
  return true;
}

//check if autofocus is currently running
uint8_t getAutofocusStatus()
{
  return camera->get_reg(camera, OV5640_CMD_FW_STATUS, 0xff);
}

//perform a single autofocus
void triggerAutofocus()
{
  sendAutofocusCommandNoAck(AF_TRIG_SINGLE_AUTO_FOCUS);
}

//turn on continuous autofocus, which will automatically perform an autofocus every few seconds
void enableContinuousAutofocus()
{
  sendAutofocusCommandNoAck(AF_CONTINUE_AUTO_FOCUS);
}

//manually set focus position. 0 = farthest, 255 = nearest
void setManualFocus(uint8_t position)
{
  camera->set_reg(camera, OV5640_CMD_PARA3, 0xff, 0x00);
  camera->set_reg(camera, OV5640_CMD_PARA4, 0xff, position);
  sendAutofocusCommandNoAck(AF_SET_MANUAL_FOCUS);
}

//read the current focus position. 0 = farthest, 255 = nearest
byte getFocusPosition()
{
  sendAutofocusCommand(AF_GET_FOCUS_POSITION);
  return camera->get_reg(camera, OV5640_CMD_PARA4, 0xff);
}

//grab a pixel from the camera framebuffer, convert it to an RGB565 color, and perform heckery to it
uint16_t getFramebufferColor(uint32_t addr)
{
  addr *= 2; //each framebuffer pixel is 2 bytes
  uint16_t color = (camera_frame->buf[addr] << 8) | camera_frame->buf[addr+1]; //grab 2 bytes from camera framebuffer and combine to 16-bit uint
  color &= 0b1111011110011110; //downsample color from RGB565 to RGB444 for extra cronch
  return color;
}
/*
const uint16_t palette_lc[16] = {0xD590, 0xC3CB, 0xAAE8, 0x7A29, 0x49E8, 0xBC8B, 0x93A8, 0x4A27, 0x73A7, 0xB52A, 0xD654, 0x8D54, 0x4B8D, 0x524A, 0x83CE, 0xACD1};
uint16_t applyPalette(uint16_t color)
{
  uint16_t new_color = 0, error = 65535;
  int8_t r, g, b, test_r, test_g, test_b;
  splitRGB(color, r, g, b);
  for(byte i = 0; i < 16; i++)
  {
    uint16_t test_color = palette_lc[i];
    splitRGB(test_color, test_r, test_g, test_b);
    uint16_t test_error = abs(r - test_r) + abs(g - test_g) + abs(b - test_b);
    if(test_error < error)
    {
      error = test_error;
      new_color = test_color;
    }
  }
  return new_color;
}
void splitRGB(uint16_t color, int8_t &r, int8_t &g, int8_t &b)
{
  r = (color >> 11) & 0b11111;
  g = (color >> 5) & 0b111111;
  b = color & 0b11111;
}*/