byte setupSDCard()
{
  if(!cardInSlot()) { return false; }
  //try to initialize SD card
  if(!sd.begin(SD_CS, SD_SCK_MHZ(4)))
  {
    if(sd.card()->errorCode()) { showStatus(bmp_sd_error, "sd card fault", COLOR_INDICATOR_NG); }
    else if(sd.vol()->fatType() == 0) { showStatus(bmp_sd_error, "filesystem incompatible", COLOR_INDICATOR_NG); }
    else { showStatus(bmp_sd_error, "sd card fault", COLOR_INDICATOR_NG); }
    beep(TONE_ERROR);
    return false;
  }
  //check what the first available file number is
  if(!findNextFilename())
  {
    showStatus(bmp_sd_error, "too many files", COLOR_INDICATOR_NG);
    beep(TONE_ERROR);
    return false;
  }
  return true;
}

//returns 1 if an SD card is physically inserted in the slot (uses a switch in the slot)
byte cardInSlot()
{
  return aw.digitalRead(AWEXP_SD_DET);
}

//checks for the next available filename in the BLS-####.BMP format
byte findNextFilename()
{
  while(photo_number <= 9999)
  {
    sprintf(photo_filename, "BLS-%04d.PNG", photo_number);
    if(!sd.exists(photo_filename)) { return true; }
    photo_number++;
  }
  return false;
}

#define PNG_COMPRESS_LEVEL 3 //1 = fastest, 9 = slowest
uint8_t png_file_buffer[102400]; //buffer for entire PNG file
uint8_t png_888_buffer[320 * 3]; //line buffer for converting RGB565 pixels to RGB888
void takePhoto()
{
  beep(TONE_CAPTURE);
  setNeopixel(50, 20, 40);
  showStatus(bmp_sd_saving, "recording...", COLOR_INDICATOR_OK);
  drawStatus();
  updateDisplay();
  
  //first, downsample camera frame colors from RGB565 to RGB444 for extra cronch
  uint32_t num_pixels = camera_frame->width * camera_frame->height;
  for(uint32_t i = 0; i < num_pixels; i++)
  {
    uint16_t color = getFramebufferColor(i);
    camera_frame->buf[i*2+1] = color >> 8;
    camera_frame->buf[i*2] = color & 0xff;
  }
  
  //encode camera frame into png
  png.open(png_file_buffer, sizeof(png_file_buffer));
  png.encodeBegin(camera_frame->width, camera_frame->height, PNG_PIXEL_TRUECOLOR, 24, NULL, PNG_COMPRESS_LEVEL); //24-bit truecolor encoding
  for(int16_t line = 0; line < camera_frame->height; line++)
  {
    png.addRGB565Line((uint16_t*)(&camera_frame->buf[line * camera_frame->width * 2]), png_888_buffer);
  }

  //try to write png file to sd card
  if(png.getLastError())
  {
    showStatus(bmp_sd_error, "png encode failed", COLOR_INDICATOR_NG);
    beep(TONE_ERROR);
  }
  else
  {
    uint32_t filesize = png.close();
    File file;
    if(file.open(photo_filename, FILE_WRITE) && file.write(png_file_buffer, filesize))
    {
      findNextFilename();
      clearStatus();
      beep(TONE_SAVED);
    }
    else
    {
      showStatus(bmp_sd_error, "save failed", COLOR_INDICATOR_NG);
      sdcard_ready = 0;
      beep(TONE_ERROR);
    }
    file.close();
  }
  setNeopixel(0, 0, 0);

  //save parameters whenever a photo is taken
  saveParams();
}