uint32_t sdcard_insert_time = 0;
void loop()
{
  //if an SD card was just inserted, wait a second, then try to initialize it
  if(justReleased(AWEXP_SD_DET))
  {
    sdcard_insert_time = millis() + 1000;
    showStatus(bmp_sd_saving, "sd card access", COLOR_INDICATOR_OK);
  }
  if(sdcard_insert_time && millis() >= sdcard_insert_time)
  {
    sdcard_insert_time = 0;
    clearStatus();
    sdcard_ready = setupSDCard();
  }
  if(!cardInSlot())
  {
    sdcard_insert_time = 0;
    sdcard_ready = 0;
    showStatus(bmp_sd_missing, "sd card not inserted", COLOR_INDICATOR_NG);
  }
  camera_frame = esp_camera_fb_get(); //capture frame from camera
  blink_register = !blink_register; //this variable flips every frame for use with blinking UI elements
  drawViewfinderUI();
  readButtons();
  if(optionsmenu_open) { drawOptionsMenu(); }
  else { handleButtons(); }
  updateDisplay();

  //update focus position value if autofocus in progress or just completed
  byte focus_status = isInAuto(PARAM_FOCUS);
  if(focus_status || (!focus_status && old_focus_status)) { getUIValue(PARAM_FOCUS); }
  old_focus_status = focus_status;

  //if shutter AF is enabled and autofocus just completed, take a photo
  if(!focus_status && shutter_af_waiting)
  {
    shutter_af_waiting = 0;
    takePhoto();
  }

  //save parameters if any parameters haven't been adjusted for a little while
  if(param_save_time && millis() >= param_save_time) { saveParams(); }

  esp_camera_fb_return(camera_frame); //return framebuffer to camera driver so another frame can be captured
}

void handleButtons()
{
	int8_t modify_amt = isPressed(AWEXP_BUTTON_UP) + isPressed(AWEXP_BUTTON_DOWN) * -1;
	if(modify_amt != 0)
	{
    incrementUIValue(cursor_pos, modify_amt);
    if(cursor_pos == PARAM_FOCUS) { focus_magnifier_end_time = millis() + FOCUS_MAGNIFIER_TIME; } //enable focus magnifier if focus was manually adjusted
    armParamSaveTimer(cursor_pos); //if a flash-saved parameter was adjusted, start timer for saving it to flash
	}
	else if(justPressed(AWEXP_BUTTON_RIGHT))
  {
    cursor_pos--;
    if(cursor_pos < 0) { cursor_pos = NUM_PARAMS - 1; }
  }
	else if(justPressed(AWEXP_BUTTON_LEFT))
  {
    cursor_pos++;
    if(cursor_pos >= NUM_PARAMS) { cursor_pos = 0; }
  }
	else if(justPressed(AWEXP_BUTTON_SEL))
	{
    reset(cursor_pos);
    armParamSaveTimer(cursor_pos); //if a flash-saved parameter was adjusted, start timer for saving it to flash
	}
  else if(justPressed(AWEXP_BUTTON_OK))
  {
    optionsmenu_open = 1;
    optionsmenu_open_time = millis();
  }
  else if(justPressed(SHUTTER_BUTTON) && sdcard_ready)
  {
    if(option_shutter_af->enabled)
    {
      triggerAutofocus();
      showStatus("focusing...");
      beep(TONE_SHUTTERAF);
      shutter_af_waiting = 1;
    }
    else
    {
      takePhoto();
    }
  }
}